use std::ops::Deref;

use serde::{Deserialize, Serialize};

#[derive(Debug, Clone, Serialize, Deserialize, PartialEq, Eq)]
pub struct AudioMetadata(pub serde_json::Value);

impl AudioMetadata {
    /// Get a key from the JSON data if it exists.
    fn get_key(&self, key: &str) -> Option<&str> {
        self.0.as_object()?.get(key)?.as_str()
    }

    /// Get the title of the audio metadata if it exists.
    pub fn title(&self) -> Option<&str> {
        self.get_key("title")
    }

    /// Get the artist of the audio metadata if it exists.
    pub fn artist(&self) -> Option<&str> {
        self.get_key("artist")
    }

    /// Get the album of the audio metadata if it exists.
    pub fn album(&self) -> Option<&str> {
        self.get_key("album")
    }

    /// Get the track number of the audio metadata if it exists.
    pub fn track_number(&self) -> Option<usize> {
        // Note: This will return an error if the track number is not a valid integer
        self.get_key("tracknumber").map(|x| x.parse().ok())?
    }
}

/// Get the audio metadata for a given file path, or None if it fails.
pub fn get_metadata(file_path: &str) -> Option<AudioMetadata> {
    // Simply wrap the JSON data in an `AudioMetadata` instance
    Some(AudioMetadata(get_metadata_json(file_path)?))
}

/// Get the JSON data for a given file path using the `extract_metadata.py` script, or None if it fails.
pub fn get_metadata_json(file_path: &str) -> Option<serde_json::Value> {
    let output = std::process::Command::new("python3")
        .arg("./extract_metadata.py")
        .arg(file_path)
        .output()
        .ok()?;

    let stdout = String::from_utf8_lossy(&output.stdout);
    serde_json::from_str(&stdout).ok()?
}

impl Deref for AudioMetadata {
    type Target = serde_json::Value;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}
