use album::Album;
use artist::Artist;
use based::{get_pg, request::cache::RouteCache};
use serde_json::json;
use std::{
    path::{Path, PathBuf},
    str::FromStr,
};
use track::Track;
use walkdir::WalkDir;
pub mod album;
pub mod artist;
pub mod event;
pub mod metadata;
pub mod playlist;
pub mod search;
pub mod track;

/// Checks if a file has a music file extension
fn is_music_file(path: &Path) -> bool {
    if let Some(extension) = path.extension() {
        match extension.to_str().unwrap_or("").to_lowercase().as_str() {
            "flac" | "wav" | "aac" | "ogg" | "opus" => return true,
            _ => return false,
        }
    }
    false
}

pub struct Libary {
    root_dir: PathBuf,
}
impl Libary {
    /// Creates a new instance of the Libary struct with the specified `root_dir`.
    pub const fn new(root_dir: PathBuf) -> Self {
        Self { root_dir }
    }

    /// Finds or creates an artist by their name and returns its ID.
    ///
    /// If the artist already exists, this method returns their ID. Otherwise, it creates a new artist record and returns its ID.
    pub async fn find_or_create_artist(&self, artist: &str) -> uuid::Uuid {
        if let Some(artist) = Artist::find(artist).await {
            // Return the existing artist's ID.
            artist.id
        } else {
            // Create a new artist record with the given name and return its ID.
            Artist::create(artist).await.id
        }
    }

    /// Finds or creates an album by its title and returns its ID.
    ///
    /// This method takes an optional `artist_id` parameter, which specifies the
    /// artist to associate with the new album. If no `artist_id` is provided,
    /// the album will be created without an associated artist.
    pub async fn find_or_create_album(
        &self,
        album: &str,
        artist_id: Option<uuid::Uuid>,
    ) -> uuid::Uuid {
        if let Some(album) = Album::find(album, artist_id).await {
            // Return the existing album's ID.
            album.id
        } else {
            // Create a new album record with the given title and return its ID.
            Album::create(album, artist_id).await.id
        }
    }

    /// Adds a music file at the specified `path` to the library.
    ///
    /// This method searches for the path in the existing library records. If it's
    /// already present, no action is taken. Otherwise, it extracts metadata from
    /// the file and creates a new track record with associated artist and album IDs,
    /// if available.
    pub async fn add_path_to_library(&self, path: &str) {
        // Check if the path is already present in the library records.
        if Track::of_path(path).await.is_some() {
            return;
        }

        log::info!("Adding {path} to library");

        // Extract metadata from the file at the specified path.
        let metadata = metadata::get_metadata(path);

        let mut entry = json!({
            "path": path,
        });

        if let Some(meta) = &metadata {
            // If artist information is available, create or retrieve its ID and associate it with this track record.
            if let Some(artist) = meta.artist() {
                let artist_id = self.find_or_create_artist(artist).await;
                entry
                    .as_object_mut()
                    .unwrap()
                    .insert("artist".into(), artist_id.to_string().into());
            } else {
                log::warn!("{path} has no artist");
            }

            // If album information is available, create or retrieve its ID and associate it with this track record.
            if let Some(album) = meta.album() {
                let album_id = self
                    .find_or_create_album(
                        album,
                        entry
                            .as_object()
                            .unwrap()
                            .get("artist")
                            .map(|x| uuid::Uuid::from_str(x.as_str().unwrap()).unwrap()),
                    )
                    .await;

                entry
                    .as_object_mut()
                    .unwrap()
                    .insert("album".into(), album_id.to_string().into());
            } else {
                log::warn!("{path} has no album and will be treated as single");
            }

            // Extract the title of this track from the metadata.
            if let Some(title) = meta.title() {
                entry
                    .as_object_mut()
                    .unwrap()
                    .insert("title".into(), title.into());
            }
        }

        // Add any available metadata to the track record.
        if let Some(metadata) = metadata {
            entry
                .as_object_mut()
                .unwrap()
                .insert("meta".into(), metadata.0);
        }

        // If no title is specified in the metadata, use the file name as the title.
        if entry.as_object().unwrap().get("title").is_none() {
            entry.as_object_mut().unwrap().insert(
                "title".into(),
                std::path::Path::new(path)
                    .file_stem()
                    .unwrap()
                    .to_str()
                    .unwrap()
                    .into(),
            );
        }

        // Create a new track record with the extracted metadata and add it to the library.
        Track::create(entry.as_object().unwrap()).await;
    }

    /// Retrieves all artists from the database.
    pub async fn get_artists(&self) -> Vec<Artist> {
        sqlx::query_as("SELECT * FROM artist")
            .fetch_all(get_pg!())
            .await
            .unwrap()
    }

    /// Retrieves all albums for a given artist ID.
    pub async fn get_albums_by_artist(&self, artist: &uuid::Uuid) -> Vec<Album> {
        sqlx::query_as("SELECT * FROM album WHERE artist = $1")
            .bind(artist)
            .fetch_all(get_pg!())
            .await
            .unwrap()
    }

    /// Retrieves all singles (tracks with no album) for a given artist ID.
    pub async fn get_singles_by_artist(&self, artist: &uuid::Uuid) -> Vec<Track> {
        sqlx::query_as("SELECT * FROM track WHERE album IS NULL AND artist = $1")
            .bind(artist)
            .fetch_all(get_pg!())
            .await
            .unwrap()
    }

    /// Retrieves an album record by its ID, returning None if not found.
    pub async fn get_album_by_id(&self, album: &uuid::Uuid) -> Option<Album> {
        Album::get(album).await
    }

    /// Retrieves a track record by its ID, returning None if not found.
    pub async fn get_track_by_id(&self, track_id: &uuid::Uuid) -> Option<Track> {
        Track::get(track_id).await
    }
    /// Reloads metadata for a given track ID.
    ///
    /// This method retrieves the current metadata associated with the track,
    /// updates it if necessary (e.g., artist or album is missing), and then
    /// persists the updated metadata to the database.
    pub async fn reload_metadata(&self, track_id: &uuid::Uuid) -> Result<(), ()> {
        // Retrieve the track record from the database.
        let track = Track::get(track_id).await.ok_or_else(|| ())?;

        // Get the path associated with the track (e.g., file system location).
        let path = &track.path;

        log::info!("Rescanning metadata for {path}");

        // Retrieve the current metadata for the track.
        let metadata = metadata::get_metadata(path);

        // Create an update object to store changes to the metadata.
        let mut update = json!({});

        // If metadata is available, extract artist and album information.
        if let Some(meta) = &metadata {
            // Get the artist associated with the track (if present).
            if let Some(artist) = meta.artist() {
                // Create or retrieve an artist record in the database.
                let artist_id = self.find_or_create_artist(artist).await;

                // Add the artist ID to the update object.
                update
                    .as_object_mut()
                    .unwrap()
                    .insert("artist".into(), artist_id.to_string().into());
            } else {
                log::warn!("{path} has no artist");
            }

            // Get the album associated with the track (if present).
            if let Some(album) = meta.album() {
                // Create or retrieve an album record in the database, using
                // the update object to pass additional metadata information.
                let album_id = self
                    .find_or_create_album(
                        album,
                        update
                            .as_object()
                            .unwrap()
                            .get("artist")
                            .map(|x| uuid::Uuid::from_str(x.as_str().unwrap()).unwrap()),
                    )
                    .await;

                // Add the album ID to the update object.
                update
                    .as_object_mut()
                    .unwrap()
                    .insert("album".into(), album_id.to_string().into());
            } else {
                log::warn!("{path} has no album and will be treated as single");
            }

            // Get the track title (if present).
            if let Some(title) = meta.title() {
                update
                    .as_object_mut()
                    .unwrap()
                    .insert("title".into(), title.into());
            }
        }

        // Add any remaining metadata to the update object.
        if let Some(metadata) = metadata {
            update
                .as_object_mut()
                .unwrap()
                .insert("meta".into(), metadata.0);
        }

        // If no track title is available, use the file name as a fallback.
        if update.as_object().unwrap().get("title").is_none() {
            update.as_object_mut().unwrap().insert(
                "title".into(),
                std::path::Path::new(&path)
                    .file_stem()
                    .unwrap()
                    .to_str()
                    .unwrap()
                    .into(),
            );
        }

        // Apply the updated metadata to the track record.
        track.update(&update).await.unwrap();

        Ok(())
    }

    /// Cleans up lost files from the database.
    ///
    /// This method iterates over all tracks, albums, and artists in the
    /// database and removes any records that no longer exist on disk.
    pub async fn clean_lost_files(&self) {
        // Clean up lost track records.
        for track in Track::find_all().await {
            if !std::path::Path::new(&track.path).exists() {
                log::info!("Cleaning lost {}", track.path);
                track.remove().await;
            }
        }

        // Clean up lost album records.
        for album in Album::find_all().await {
            if Track::find_of_album(&album.id).await.is_empty() {
                log::info!("Cleaning album {} with no tracks", album.title);
                album.remove().await;
            }
        }

        // Clean up lost artist records.
        for artist in Artist::find_all().await {
            if Track::find_first_of_artist(&artist.id).await.is_none()
                && Album::find_of_artist(&artist.id).await.is_empty()
            {
                log::info!("Cleaning artist {} with no tracks or albums", artist.name);
                artist.remove().await;
            }
        }
    }

    /// Rescan the entire library.
    pub async fn rescan(&self, cache: &RouteCache) {
        cache.invalidate("albums", "latest").await;

        log::info!("Rescanning library");

        for entry in WalkDir::new(self.root_dir.clone())
            .follow_links(true)
            .into_iter()
            .filter_map(std::result::Result::ok)
        {
            let path = entry.path();

            if path.is_file() && is_music_file(path) {
                let path = path.to_string_lossy().to_string();

                self.add_path_to_library(&path).await;
            }
        }
    }
}
