use crate::get_pg;
use crate::library::playlist::Playlist;
use crate::library::playlist::Visibility;
use crate::library::track::Track;
use based::auth::User;
use based::request::api::api_error;
use based::request::api::to_uuid;
use based::request::api::vec_to_api;
use based::request::api::FallibleApiResponse;
use rocket::get;
use rocket::post;
use rocket::serde::json::Json;
use serde_json::json;

#[get("/playlists")]
pub async fn playlists_route(u: User) -> FallibleApiResponse {
    let mut playlists = vec![
        json!({"id": "recents", "name": "Recently Played"}),
        json!({"id": "recentlyAdded", "name": "Recently Added"}),
    ];

    let own_playlists = Playlist::of(&u).await;

    for playlist in own_playlists {
        playlists.push(json!({
            "id": playlist.id,
            "name": playlist.title
        }));
    }

    Ok(json!(playlists))
}

pub async fn recently_added_playlist() -> FallibleApiResponse {
    let tracks = Track::find_recently_added().await;
    Ok(json!(vec_to_api(&tracks).await))
}

#[get("/playlist/<id>")]
pub async fn playlist_route(id: &str, u: User) -> FallibleApiResponse {
    if id == "recents" {
        return Ok(Playlist {
            id: uuid::Uuid::nil(),
            id_str: Some("recents".to_string()),
            owner: u.username,
            title: "Recently Played".to_string(),
            visibility: Visibility::Public,
            tracks: vec![],
        }
        .api()
        .await);
    }

    if id == "recentlyAdded" {
        return Ok(Playlist {
            id: uuid::Uuid::nil(),
            id_str: Some("recentlyAdded".to_string()),
            owner: u.username,
            title: "Recently Added".to_string(),
            visibility: Visibility::Public,
            tracks: vec![],
        }
        .api()
        .await);
    }

    let playlist = Playlist::get(&to_uuid(id)?)
        .await
        .ok_or_else(|| api_error("No playlist with that ID found"))?;

    if matches!(playlist.visibility, Visibility::Private) && u.username != playlist.owner {
        return Err(api_error("Forbidden"));
    }

    Ok(playlist.api().await)
}

#[get("/playlist/<id>/tracks")]
pub async fn playlist_tracks_route(id: &str, u: User) -> FallibleApiResponse {
    if id == "recents" {
        let tracks = Track::get_latest_of_user(&u).await;
        return Ok(json!(vec_to_api(&tracks).await));
    }

    if id == "recentlyAdded" {
        return recently_added_playlist().await;
    }

    let playlist = Playlist::get(&to_uuid(id)?)
        .await
        .ok_or_else(|| api_error("No playlist with that ID found"))?;

    if matches!(playlist.visibility, Visibility::Private) && u.username != playlist.owner {
        return Err(api_error("Forbidden"));
    }

    let mut tracks: Vec<uuid::Uuid> = vec![];

    for track in playlist.tracks {
        tracks.push(track);
    }

    let tracks: Vec<Track> = sqlx::query_as("SELECT * FROM track WHERE id = ANY($1)")
        .bind(tracks)
        .fetch_all(get_pg!())
        .await
        .unwrap();

    Ok(json!(vec_to_api(&tracks).await))
}

#[derive(rocket::serde::Deserialize)]
pub struct PlaylistData {
    pub title: Option<String>,
    pub visibility: Option<Visibility>,
    pub tracks: Vec<uuid::Uuid>,
}

#[post("/playlist", data = "<playlist>")]
pub async fn playlist_add_route(playlist: Json<PlaylistData>, u: User) -> FallibleApiResponse {
    let playlist = Playlist::create(
        u,
        &playlist
            .title
            .clone()
            .ok_or_else(|| api_error("Bad Request"))?,
        playlist
            .visibility
            .clone()
            .ok_or_else(|| api_error("Bad Request"))?,
        &playlist.tracks,
    )
    .await
    .ok_or_else(|| api_error("Failed to create playlist"))?;

    Ok(json!({"created": playlist.id}))
}

#[post("/playlist/<id>", data = "<edit>")]
pub async fn playlist_edit_route(
    id: &str,
    edit: Json<PlaylistData>,
    u: User,
) -> FallibleApiResponse {
    let playlist = Playlist::get(&to_uuid(id)?)
        .await
        .ok_or_else(|| api_error("No playlist with that ID found"))?;

    if playlist.owner != u.username {
        return Err(api_error("Forbidden"));
    }

    let mut tracks_ref = vec![];

    for track in &edit.tracks {
        tracks_ref.push(
            Track::get(track)
                .await
                .ok_or_else(|| api_error("Invalid tracks found"))?
                .id,
        );
    }

    let playlist_id = playlist.id.clone();

    let new_title = edit.title.as_ref().unwrap_or(&playlist.title);
    let new_vis = edit.visibility.as_ref().unwrap_or(&playlist.visibility);

    sqlx::query("UPDATE playlist SET title = $1, visibility = $2, tracks = $3 WHERE id = $4")
        .bind(new_title)
        .bind(new_vis)
        .bind(tracks_ref)
        .bind(&playlist_id)
        .fetch_one(get_pg!())
        .await
        .unwrap();

    Ok(json!({"edited": playlist_id}))
}
